//********************************************************************************************************//
// SMARTCARD
#include <Smartcard.h>

#define SC_C2_RST 7
#define SC_C1_VCC 6
#define SC_C7_IO 10
#define SC_C3_CLK 9

// HIGH = NO CARD
// LOW = CARD INSERTED
#define SC_SWITCH_CARD_PRESENT 8
#define SC_SWITCH_CARD_PRESENT_INVERT true

SmartCardReader sc(SC_C7_IO, SC_C2_RST, SC_C1_VCC, SC_SWITCH_CARD_PRESENT, SC_C3_CLK, SC_SWITCH_CARD_PRESENT_INVERT);



//************************************************************************************************************************************//
//
void updateSmartCard()
{
  uint8_t bytes_received = 0;
  uint8_t b = 0;

  // Wait for card to be inserted into smart card slot
  if (!sc.cardInserted() && SMARTCARD_IS_WORKING)
  {
    debug_println("= [SMARTCARD] Waiting for Smartcard");
    SMARTCARD_IS_WORKING = false;
  }
  else
  {
    if (SMARTCARD_IS_WORKING) 
    {
      debug_println(F("= [SMARTCARD] Smartcard found"));
      debug_println(F("= [SMARTCARD] DISBALE I2C INTERRUPTS"));
      stopI2C();
      
      delay(1500);
      
      SMARTCARD_IS_WORKING = false;

      display.clear();
      display.string("Smartcarding");

      
      
      // CURENTLY CLK_4MHZ MIGHT LEAD TO PROBLEMS RECEIVING SMARTCARD_DATA
      
      //************************************************************************************************************************************//
      // 1. ATR - COLD RESET
      //
      bytes_received = sc.activate(SMARTCARD_DATA, MAX_ATR_BYTES, CLK_1MHZ);
      if ( bytes_received > 0)
      {
        debug_print(F("= [SMARTCARD] COLD ATR - Number of bytes received: "));
        debug_println(bytes_received);
        
        debug_println(F("= [SMARTCARD] COLD ATR - Received ATR ..."));
        sc.dumpHEX(SMARTCARD_DATA, bytes_received);



        //************************************************************************************************************************************//
        // 2. ATR - WARM RESET
        //
        bytes_received = sc.activate(SMARTCARD_DATA, MAX_ATR_BYTES, CLK_1MHZ);
        if ( bytes_received > 0)
        {
          debug_print(F("= [SMARTCARD] WARM ATR - Number of bytes received: "));
          debug_println(bytes_received);
          
          debug_println(F("= [SMARTCARD] WARM ATR - Received ATR ..."));
          sc.dumpHEX(SMARTCARD_DATA, bytes_received);



          //************************************************************************************************************************************//
          // SELECT APPLICATION
          //
          // TODO - DOES THIS CHANGE?
          // RID + AppID = apd 
          // APDU 00 A4 04 00 08 6F 68 6D 69 6E 61 70 64 00
          //
          delay(100);
          b = 0;
          SMARTCARD_DATA[b++] = 0x00;
          SMARTCARD_DATA[b++] = 0xA4;
          SMARTCARD_DATA[b++] = 0x04;
          SMARTCARD_DATA[b++] = 0x00;
          SMARTCARD_DATA[b++] = 0x08;
          SMARTCARD_DATA[b++] = 0x6F;
          SMARTCARD_DATA[b++] = 0x68;
          SMARTCARD_DATA[b++] = 0x6D;
          SMARTCARD_DATA[b++] = 0x69;
          SMARTCARD_DATA[b++] = 0x6E;
          SMARTCARD_DATA[b++] = 0x61;
          SMARTCARD_DATA[b++] = 0x70;
          SMARTCARD_DATA[b++] = 0x64;
          SMARTCARD_DATA[b++] = 0x00;

          debug_print(F("= [SMARTCARD] SELECT APPLICATION - Sending APDU: 00 A4 04 00 08 6F 68 6D 69 6E 61 70 64 00 ... "));
          sc.sendAPDU(0x00, SMARTCARD_DATA, b);
          for (uint8_t i = 0; i < 100; i++)
          {
            bytes_received = sc.receiveBytes(SMARTCARD_DATA, 152);
            if (bytes_received > 0) break;
            debug_print(F("."));
          }
          debug_println();
          debug_print(F("= [SMARTCARD] SELECT APPLICATION - Number of bytes received: "));
          debug_print(bytes_received);
          debug_println();
          
          if (bytes_received > 0) sc.dumpHEX(SMARTCARD_DATA, bytes_received);



          //************************************************************************************************************************************//
          // VERIFY COMMAND
          //
          // PIN Input  PIN = 01 02 03 04 APDU 00 20 00 00 04 01 02 03 04
          //
          delay(100);
          b = 0;
          SMARTCARD_DATA[b++] = 0x00;
          SMARTCARD_DATA[b++] = 0x20;
          SMARTCARD_DATA[b++] = 0x00;
          SMARTCARD_DATA[b++] = 0x00;
          SMARTCARD_DATA[b++] = 0x04;
          // PIN
          SMARTCARD_DATA[b++] = PIN_HEX[0];
          SMARTCARD_DATA[b++] = PIN_HEX[1];
          SMARTCARD_DATA[b++] = PIN_HEX[2];
          SMARTCARD_DATA[b++] = PIN_HEX[3];

          debug_print(F("= [SMARTCARD] VERIFY COMMAND - Sending APDU: 00 20 00 00 04 XX XX XX XX ... "));
          sc.sendAPDU(0x00, SMARTCARD_DATA, b);

          for (uint8_t i = 0; i < 100; i++)
          {
            bytes_received = sc.receiveBytes(SMARTCARD_DATA, 152);
            if (bytes_received > 0) break;
            debug_print(F("."));
          }
          debug_println();
          debug_print(F("= [SMARTCARD] VERIFY COMMAND - Number of bytes received: "));
          debug_print(bytes_received);
          debug_println();
          
          if (bytes_received > 0) sc.dumpHEX(SMARTCARD_DATA, bytes_received);
          
          
          
          //************************************************************************************************************************************//
          // PIN CORRECT
          // CORRECT = 0x90
          //
          if((uint8_t)SMARTCARD_DATA[3] == 144)
          {
            Serial.println("= [SMARTCARD] VERIFY COMMAND - CORRECT PIN!");
            display.clear();
            display.string("PIN Correct");
          }
          
          
          
          //************************************************************************************************************************************//
          // PIN WRONG 
          // WRONG = 0x63, NUMBER TRIES LEFT = 0xC2
          //
          if((uint8_t)SMARTCARD_DATA[3] == 99)
          {
            Serial.println("= [SMARTCARD] VERIFY COMMAND - WRONG PIN!!!!!");
            display.clear();
            display.string("PIN Wrong! ");
            display.string(" Press KEY to enter PIN");
            
            // CHANGE STATE - WE HAVE FAILED AND NEED TO WAIT FOR EITHER JOLLA OR A CORRECT PIN
            ARDUINO_IS_WORKING = true;
            sc.deactivate();
            
            debug_println(F("= [SMARTCARD] ENABLE I2C INTERRUPTS..."));
            startI2C();
            return;
          }



          //************************************************************************************************************************************//
          // SECURITY OPERATION COMMAND
          //
          // COMPUTE DIGITAL SIGNATUR Input
          // APDU 00 2A 9E 9A 05 68 61 6C 6C 6F 00
          //
          delay(100);
          b = 0;
          SMARTCARD_DATA[b++] = 0x00;
          SMARTCARD_DATA[b++] = 0x2A;
          SMARTCARD_DATA[b++] = 0x9E;
          SMARTCARD_DATA[b++] = 0x9A;
          
          // ADD LENGTH
          // QUICK FIX FORCE MAX 25BYTE LENGTH MESSAGE
          if(MESSAGE_TO_ENCRYPT_SIZE > 25)
          {
            MESSAGE_TO_ENCRYPT_SIZE = 25;  // CUT THE REST MAINLY THE AMOUNT VALUE SO ONLY ID AND IBAN
          }
          
          SMARTCARD_DATA[b++] = MESSAGE_TO_ENCRYPT_SIZE;
          debug_print(F("= [SMARTCARD] COMPUTE DIGITAL SIGNATUR - MESSAGE: "));
          debug_print(MESSAGE_TO_ENCRYPT);
          debug_print(F(" - Sending APDU: 00 2A 9E 9A "));
          debug_printHEX(MESSAGE_TO_ENCRYPT_SIZE, HEX);
          debug_print(F(" "));
          
          // ADD VALUES
          for(uint8_t m = 0; m < MESSAGE_TO_ENCRYPT_SIZE; m++)
          {
            SMARTCARD_DATA[b++] = MESSAGE_TO_ENCRYPT[m];
            debug_printHEX(MESSAGE_TO_ENCRYPT[m], HEX);
            debug_print(" ");
          }

          SMARTCARD_DATA[b++] = 0x00;
          debug_print(F("0x00"));
          
          sc.sendAPDU(0x00, SMARTCARD_DATA, b);
          
          for (uint8_t i = 0; i < 100; i++)
          {
            bytes_received = sc.receiveBytes(SMARTCARD_DATA, 152);
            if (bytes_received > 0) break;
            debug_print(F("."));
            if (i % 100 == 0) debug_println();
          }
          debug_println();
          debug_print(F("= [SMARTCARD] COMPUTE DIGITAL SIGNATUR - Number of bytes received: "));
          debug_println(bytes_received);
          debug_println();
          
          if (bytes_received > 0) sc.dumpHEX(SMARTCARD_DATA, bytes_received);

          debug_println(F("= [SMARTCARD] COMPUTE DIGITAL SIGNATUR - Waiting for SmartCard to finish Security Operation Command!"));
          delay(2000);
          debug_println(F("= [SMARTCARD] COMPUTE DIGITAL SIGNATUR - Finished Security Operation Command!"));
          
          
          
          //************************************************************************************************************************************//
          // FINSHED
          // CHANGE STATE - WE HAVE FINISHED SUCCESSUFULLY AND CAN RESPOND TO JOLLA WITH SMARTCARD_DATA
          //
          ARDUINO_IS_WORKING = false;
          display.string(" Generated   Result");
        }
      }
      else
      {
        //************************************************************************************************************************************//
        // FAILED TO READ SMARTCARD
        //
        debug_println(F("= [SMARTCARD] ERROR - Unable to identify card... Please reinsert!"));
      }
      
      //************************************************************************************************************************************//
      // RESET I2C AND DEACTIVATE SMARTCARD
      //
      delay(2000);
      sc.deactivate();
      debug_println(F("= [SMARTCARD] ENABLE I2C INTERRUPTS..."));
      startI2C();
    }
  }
}



//************************************************************************************************************************************//
// ACTIVATE TWI - I2C COMMUNICATIION
// WE NEED THIS TO DISABLE THE INTERRUPT PULLED BY JOLLA DURING THE SMARTCARD I/O PROCESS TO SECURE SMARTCARD I/O COMMUNICATION
//
void startI2C()
{
    TWCR = (1<<TWEN)|                             // TWI Interface enabled.
           (1<<TWIE)|(1<<TWINT)|                  // Enable TWI Interupt and clear the flag.
           (0<<TWEA)|(1<<TWSTA)|(0<<TWSTO)|       // Initiate a START condition.
           (0<<TWWC); 
}




//************************************************************************************************************************************//
// DEACTIVATE TWI - I2C COMMUNICATIION
// WE NEED THIS TO DISABLE THE INTERRUPT PULLED BY JOLLA DURING THE SMARTCARD I/O PROCESS TO SECURE SMARTCARD I/O COMMUNICATION
//
void stopI2C()
{
    TWCR = (1<<TWEN)|                                 // TWI Interface enabled
           (0<<TWIE)|(1<<TWINT)|                      // Disable TWI Interrupt and clear the flag
           (0<<TWEA)|(0<<TWSTA)|(1<<TWSTO)|           // Initiate a STOP condition.
           (0<<TWWC); 
}